/* $Log$ */

#ifndef NETWORK_LAYER_H
#define NETWORK_LAYER_H


/* Declarations for DLC ports attached to the network layer */
#define MAXPORT 16 

/* is the dlc port an input/output port ? */
#define OUTPUT_PORT 1
#define INPUT_PORT 0

/* Is the network layer node a source, destination or a router ? */
#define SOURCE_NODE 0
#define DEST_NODE 1
#define ROUTER_NODE 2

/* DLC port structure */
/* Every network layer will have a bunch of DLC port structures */
typedef struct _DLCPort_t { 
     Component *to_dlc;        

     /* ----- per port parameters --- */
     Param *porttype;           /* Port type: input or output. */
     Param *mtu;	    	/* MTU size  */
     Param *speed;	    	/* Transmission Speed */
     tick_t earliest_xmit_time; /* earliest transmission time */
     
} DLCPort_t;


typedef struct _NETWORKConnectiont{
    struct _NETWORKConnectiont *cn_next, *cn_prev;
    short         cn_class;
    short         cn_type;
    char          cn_name[40];    /* Name of component (appears on screen) */
    PFP           cn_action;      /* Main function of component.  */
    XComponent    *cn_picture;    /* Graphics object that displays this thing */
    list          *cn_neighbors;  /* List of neighbors of this thing */

    /* Parameters-- data that will be displayed on screen */
    short         cn_menu_up;     /* If true, then the text window is up */
    short         cn_big_menu;     /* not used */
    queuet         *cn_params;     /* Variable-length queue of parameters */

    int           cn_iq_length;    /* get the current input q length here */
    int           cn_iq_limit;    /* get the current input q length here */


    int           cn_flash;
    Param         *cn_Name;
    Param         *cn_nodetype;  /* source, destination or router ? */

     /* variables students need not be concerned with */
     int 	Nports;		  /* number of dlc ports: allows random choice */
     int 	lastport;         /* for deterministic traversal of ports */
     short 	unique_pkt_id;    /* Unique packet id for data sent
				     from this Source */

     tick_t 	cn_earliest_presentation_time; /* models presentation serialization and delay */
     tick_t 	cn_earliest_receive_time; /* models reassembly serialization and delay */
     
     int 	num_fragments;
     int	num_packets;

     int	num_packets_timedout;
     int 	num_partially_reassembled;
     int 	max_partially_reassembled;
     
     /* variables of interest to students */
     DLCPort_t	dlcport[MAXPORT]; /* dlc port variables */
     struct ipfq ipfqt[IP_FQSIZE];/* Fragment lists table: for Destination  */
     
} NETWORKConnectiont;

/* ------------------- MACRO DEFINITIONS/WRAPPERS -------------- */

int num_frags_generated = 0;

#include "network_layer_wrappers.c"

#endif
  
