/* ------------- WRAPPERS FOR THE NETWORK_LAYER CODE ---------- */
/* --- Please do not modify this file. --------- */
/* --- Any doubts, send email to shivkuma@cis.ohio-state.edu
   ----- */

#define NETWORK_LAYER_ENTITY_TYPE NETWORKConnectiont
#define GENERIC_LAYER_ENTITY_TYPE Component
     

extern Log pkt_log;

static caddr_t cn_reset(),cn_create(),cn_neighbor(),
               cn_uneighbor(),cn_route(),cn_hops(),cn_send(), cn_start();

static NetworkLayerReceive(), send_pdu_to_application(),
     send_pdu_to_datalink(), choose_output_datalink(),
     NetworkToDatalink(),
     FormPacketsAndSendToDLC(), FragmentPacketAndSendToDLC(),
     NetworkToApplication();

static PDU_TYPE *CreatePacket(int , int , short , short, short , char *);

static int ipfadd();
static PDU_TYPE *ipfjoin(), *ipfcons();
void ipftimer();
static PDU_TYPE *ipreass(NETWORK_LAYER_ENTITY_TYPE *, PDU_TYPE *);


static queuet * CreateOrderedList();
static int AddToOrderedList(queuet *, PDU_TYPE *, int );
static PDU_TYPE *TraverseOrderedList(queuet *, int *);
static PDU_TYPE *DequeueFromOrderedList(queuet *);
static FreeOrderedList(queuet *);

extern double random_range();

extern char *sim_calloc(), *sim_malloc();

#define IP_FRAG_TIMER (Evtype)(EV_CLASS_PRIVATE | 1)
#define MY_RECEIVE  (Evtype)(EV_CLASS_PRIVATE | 6)

/****************************************/
#define NETWORK_LAYER_ENTITY_TYPE NETWORKConnectiont
#define GENERIC_LAYER_ENTITY_TYPE Component

#define NetworkFromApplication(src) ((src)->co_class == APP_CLASS)
#define NetworkFromDatalink(src) ((src)->co_class == DLC_CLASS)

#define SourceEndSystem(nw) ((nw)->cn_nodetype->u.i == SOURCE_NODE)
#define DestinationEndSystem(nw) ((nw)->cn_nodetype->u.i == DEST_NODE)
#define RouterNode(nw) ((nw)->cn_nodetype->u.i == ROUTER_NODE)

#define DropPDU(pdu) pdu_free(pdu)
#define GetMTU(cn,i) cn->dlcport[i].mtu->u.i
#define GetPktId(cn) cn->unique_pkt_id
#define UpdatePktId(cn) cn->unique_pkt_id++


/****************************************/

caddr_t network_layer_action(Component *src,
			 NETWORKConnectiont *cn,
			 int type,
			 PDU_TYPE *pdu,
			 caddr_t arg)
{
     caddr_t result = NULL;	      			      
     switch (type)  {
     case EV_RECEIVE:{
	  if (DestinationEndSystem(cn)){
               /* 20 usecs processing delay */
	       tick_t proc_delay = USECS_TO_TICKS(20);
	       tick_t t = Maxm( cn->cn_earliest_receive_time , ev_now());

	       /* fake another receive from the same source after
		  the queueing delay */
	       ev_enqueue(MY_RECEIVE, (Component *)src, (Component *)cn,
			  t + proc_delay,
			  cn->cn_action, (PDU_TYPE *) pdu,
			  (caddr_t)NULL);

	       cn->cn_earliest_receive_time = t + proc_delay;
	  }
	  else{
	       NetworkLayerReceive(cn,src,pdu);
	  }
	  result = (caddr_t)cn;
	  break;
     }
     case MY_RECEIVE: /* destination special: models reassembly
			 and serialization delays */ 
	  NetworkLayerReceive(cn,src,pdu);
	  result = (caddr_t)cn;
	  break;	
     case EV_RESET:
     {
	  cn->cn_iq_length = 0;
	  cn->cn_iq_limit = 1;
	  result=(caddr_t)cn;
	  break;
     }
     case EV_CREATE:
	  result = cn_create((char *)arg);
	  break;
     case EV_DEL:
	  comp_delete((Component *)cn);
	  result = (caddr_t)cn;
	  break;
     case EV_NEIGHBOR:
	  result = cn_neighbor(cn, (Component *)arg);
	  break; 
     case EV_UNEIGHBOR:
	  result = cn_uneighbor(cn, (Component *)arg);
	  break;
     case EV_LEGAL_NEXT_HOPS:
	  result = cn_hops(cn);
	  break;
     case EV_MAKE_ROUTE:
	  result = cn_route(cn, (list *)arg);
	  break;
     case EV_START:
	  result = cn_start(cn); /* _todo_ :  cn_start(cn): sets frag timer */
	  break;
     case IP_FRAG_TIMER:{
	  ipftimer(cn);
	  ev_enqueue(IP_FRAG_TIMER, (Component *)cn, 
		     (Component *)cn,
	     ev_now()+USECS_TO_TICKS(FTTL_DECR_USECS),
		     network_layer_action, (PDU_TYPE *)NULL,(caddr_t)NULL);
	  result = (caddr_t)cn;
	  break;
     }
     case EV_UPDATE_ROUTE_ACTIVITY:
	  update_routes(src,(Component *)cn);
	  result=(caddr_t)cn;
	  break;
     default:
	  break;			  
     }					    
     return(result);			      
}
/****************************************/

static
send_pdu_to_datalink(NETWORKConnectiont *cn, PDU_TYPE *pdu, int i)
{ /* choice of o/p link, fragmentation etc already done */
Component *c = cn->dlcport[i].to_dlc;
tick_t t = Maxm( cn->dlcport[i].earliest_xmit_time , ev_now());
tick_t incr =
    USECS_TO_TICKS(((double)pdu->u.n_pdu.ip_len*8.0)/cn->dlcport[i].speed->u.d);
tick_t proc_delay = USECS_TO_TICKS(20); /* 20 usecs processing delay */

/*
     if (M.doX)
	  route_activity((Component *)cn, c, (t - ev_now())+10, 1, 2, 0);
*/
                /* unit length */


ev_enqueue(EV_RECEIVE, (Component *)cn, c, t + proc_delay,
		c->co_action, (PDU_TYPE *) pdu, (caddr_t)NULL);

     
     cn->dlcport[i].earliest_xmit_time  = t + incr + proc_delay;
     
     return 0;
}

/****************************************/
/* _todo_ */
static
send_pdu_to_application(NETWORKConnectiont *cn, PDU_TYPE *pdu)
{ /* reassembly already done */
      Neighbor *n;
      tick_t proc_delay = USECS_TO_TICKS(20); /* 20 usecs processing delay */
      tick_t t = Maxm( cn->cn_earliest_presentation_time , ev_now());

      for (n=(Neighbor *)cn->cn_neighbors->l_head;n;n=n->n_next) {
	  if (n->n_c->co_class == APP_CLASS)
	       break;
     }
      /*
     if (M.doX)
	  route_activity((Component *)cn, n->n_c, 10, 1, 2, 0);
	  */
        /* unit length app pkt */
     
/*	 route_activity((Component *)cn, n->n_c, 10, 2, 0); */

     ev_enqueue(EV_RECEIVE, (Component *)cn, n->n_c, t + proc_delay,
		n->n_c->co_action, (PDU_TYPE *) pdu,
		(caddr_t)NULL);

     cn->cn_earliest_presentation_time = t + proc_delay;
     return 0;
}

/****************************************/
static caddr_t
cn_create(name)
     register char *name;
{
    NETWORKConnectiont *newcn;
    int i;
    
   /* Memory for the component structure. */
   newcn = (NETWORKConnectiont *)sim_malloc(sizeof(NETWORKConnectiont));

   /* First things first-- copy name into the new structure. */
   strncpy(newcn->cn_name, name, 40);

   /* have to create a neighbor list */
  newcn->cn_neighbors= l_create();
  newcn->cn_params = q_create();
  newcn->cn_type = NETWORK_LAYER_TYPE;
  newcn->cn_class = NETWORK_CLASS;
  newcn->cn_action = (PFP)network_layer_action;
  newcn->cn_menu_up = FALSE;
  newcn->cn_flash = 0;

 /* Initialize the parameters */
  newcn->cn_Name = param_init ((Component*)newcn, "Name",
             (PFD)NULL, (PFP)make_name_text, (PFP)make_short_name_text,
             param_input_name,
             0, DisplayMask | InputMask | CanHaveLogMask, 0.0);

  newcn->cn_nodetype = param_init((Component *)newcn,
                    "Node Type",
                    (PFD)NULL, make_int_text, make_short_int_text,
                    param_input_int,
                    0,DisplayMask | InputMask, 0);

  newcn->cn_nodetype->u.i = SOURCE_NODE; /* default: source */

  newcn->Nports = 0;
  newcn->lastport = 0;
  newcn->unique_pkt_id = 0;
  
  for (i=0;i<MAXPORT; i++)
       newcn->dlcport[i].to_dlc = (Component *)NULL;

  return((caddr_t)newcn);
}

/****************************************/
static caddr_t
cn_neighbor(cn,c)
     register NETWORKConnectiont *cn;
     register Component *c;
{
     register Neighbor *n;
     int i;
     
     /* $Revision$: Allow unlimited number of neighbors, but at most 2 classes */
    if (!(n=add_neighbor((Component*)cn,c,-1,2,APP_CLASS,DLC_CLASS)))
        return ((caddr_t)NULL);

    if (c->co_class == DLC_CLASS) 
    {
/* $Revision$: DLC port parameters, structure which allows random choice of
   output datalink */
	 for (i=0;i<MAXPORT; i++)
	      if (!cn->dlcport[i].to_dlc) break;

	 if(i >= MAXPORT){
	      fprintf(stdout,"network_neighbor: No more output port space. Too many neighbors\n");
	      fflush(stdout);
	      exit(1);
	 }
	 cn->Nports = i+1;  /* $Revision$: _todo_: I will later choose a
			     port between 0 and Nports-1 */
	 cn->dlcport[i].to_dlc = c;
	 cn->dlcport[i].porttype = param_init((Component *)cn,
                    "Port Type of ",
                    (PFD)NULL, make_int_text, make_short_int_text,
                    param_input_int,
                    0,DisplayMask | InputMask, 0);

	 strcat(cn->dlcport[i].porttype->p_name,
		cn->dlcport[i].to_dlc->co_name);

	 cn->dlcport[i].porttype->u.i = INPUT_PORT; /* INPUT_PORT = 0 */

	 cn->dlcport[i].mtu = param_init((Component *)cn,
                    "Maximum Transmission Unit of ",
                    (PFD)NULL, make_int_text, make_short_int_text,
                    param_input_int,
                    0,DisplayMask | InputMask, 0);

	 strcat(cn->dlcport[i].mtu->p_name,
		cn->dlcport[i].to_dlc->co_name);

	 cn->dlcport[i].mtu->u.i = 576;

	 cn->dlcport[i].speed = param_init((Component *)cn,
                    "Port Transmission Speed of ",
		    (PFD)NULL, make_double_text, make_short_double_text,
                     param_input_double, 0,
	             DisplayMask | InputMask, 0.0);

	 strcat(cn->dlcport[i].speed->p_name,
		cn->dlcport[i].to_dlc->co_name);

	 cn->dlcport[i].speed->u.d = 155.52; /* in Mbps */

	 cn->dlcport[i].earliest_xmit_time = 0;
    }
    
    return ((caddr_t)n);
}

/****************************************/
static caddr_t
cn_start(NETWORKConnectiont *cn)
{
int i;

/* first order of business: set reass. timer */
     ev_enqueue(IP_FRAG_TIMER, (Component *)cn, 
		     (Component *)cn,
	     ev_now()+USECS_TO_TICKS(FTTL_DECR_USECS),/* cn->ipftimer_granularity,*/
		     network_layer_action, (PDU_TYPE *)NULL,(caddr_t)NULL);

     cn->cn_earliest_receive_time = 0;     /* used for sequentializing
					       packets to be reassembled
					       and adding a processing time */
     cn->cn_earliest_presentation_time = 0; /* used for sequentializing
					       packets to the application
					       and adding a processing time */
    cn->num_packets = cn->num_fragments = 0;
     cn->num_packets_timedout = cn->num_partially_reassembled =
	  cn->max_partially_reassembled = 0;

     
/* viewer shows only payload sizes */
     if (DestinationEndSystem(cn) && M.doX)
	  make_viewer(cn->cn_name, A_PDU_SIZE);

	 /* init ipfqt */
     for (i=0;i<IP_FQSIZE; i++){
	  cn->ipfqt[i].ipf_state = IPFF_FREE;
	  cn->ipfqt[i].nfrags = 0;/* debugging */

     }
     
     return (caddr_t)cn;
}
     
/*****************************************/
static caddr_t
cn_uneighbor(cn,c)
   register NETWORKConnectiont *cn;
   register Component *c;
{
   return ((caddr_t)remove_neighbor((Component *)cn,c));
}

/******************************************/
/* $Revision$: Junk routine: just ignore */
static caddr_t
cn_hops(cn,l)
   register NETWORKConnectiont *cn;
   list *l;
{
    register list *nlist;
    register Neighbor *n;
    register Component *c;

/* what is legal for the next hop depends on the type of the previous
   hop. If the previous one is a link, then the next must be a connection.
   If the previous one is a connection, the next one may be either a link
   or a connection.
   (Allowing a connection with both ends at the same bte ) */

   /* The previous step in the route = tail of list is put in c */
   c = (Component *)(l->l_tail->le_data);

   /* Make an empty list */
   nlist = l_create();

   /* Put appropriate neighbors into list */
   for (n=(Neighbor *)cn->cn_neighbors->l_head;n;n=n->n_next)
       if (n->n_c!=c && ((c->co_class == APP_CLASS) ||
           (c->co_class == DLC_CLASS && n->n_c->co_class == APP_CLASS)))
            l_addt(nlist,(caddr_t)n->n_c);


    return ((caddr_t)nlist);


}

/*******************************************/
/* $Revision$: Junk routine: just ignore */
static caddr_t
cn_route(cn,route_list)
         register NETWORKConnectiont *cn;
         list *route_list;
{
return ((caddr_t)cn); /* return something non-null */
}
 
/********************************************************************/

/* Fragmentation functions given to the students ... */

/****************************************/
/* Start walking on the dlc port list from an index randomly chosen between 0
   and Nports */

static 
choose_output_datalink(NETWORKConnectiont *cn)
{
     int i, initial;
     double f;
     int N;

     /* deterministic traversal */
     N = cn->Nports;
     i = initial = cn->lastport;
     cn->lastport = (cn->lastport + 1) % N;
     
     do{
	  if ( cn->dlcport[i].porttype->u.i == OUTPUT_PORT ) break;
/*	  if(initial % 2)  */
	       i = (i + 1) % (N); 
/*	  else
	       i = (i + N - 1)%(N); 
	       */
	       }
     while (i != initial);
     
/* am removing randomness for the sake of debugging */
      /* Truncate it => you get number between 0,N-1 */
	  /* traverse randomly in different directions for some fairness
	     improvement */

/*     f = random_range((double)0.0, (double) N);
     i = initial = (int) f;

     do{
	  if ( cn->dlcport[i].porttype->u.i == OUTPUT_PORT ) break;
	  if(initial % 2)
	       i = (i+1) % (N); 
	  else
	       i = (i+ N -1)%(N); 
     }
     while (i != initial);
*/


     if (cn->dlcport[i].porttype->u.i != OUTPUT_PORT){
	  fprintf(stdout,"Error: No output ports at this node\n");
	  fflush(stdout);
	  exit(1);
     }
	  
return i;
}


/* ***************************************************** */
static
FormPacketsAndSendToDLC(NETWORK_LAYER_ENTITY_TYPE *network_layer_entity,
			PDU_TYPE *input_pdu,
			int i)
{
     /* generate unique id for packet(s) */
     /* packet size depends on MTU size of connected network */
     PDU_TYPE *pdu;
     u_char *sourcep;
     int maxdlen,	/* Max data length of each fragment */
	 tosend,        /* How much *data*  remains to be sent */
	 dataoffset;    /* fragment offset within original packet */

     int N=0;
     int src = input_pdu->u.a_pdu.snode;
     int dest = input_pdu->u.a_pdu.dnode;
     short mtu = GetMTU(network_layer_entity,i);
     short pktid = GetPktId(network_layer_entity);

     if(input_pdu->type != TYPE_IS_A_PDU)
	  panic("Unknown PDU type: Expect A_PDU \n");

/* 1 conceptual IP packet (unique pkt id)  at the source split into several
   fragments immediately depending on network MTU. */

     if ((IPMHLEN + sizeof(A_PDU_TYPE)) <= mtu){ /* No fragementation required */
	  sourcep = (char *)&(input_pdu->u.a_pdu); 
	  pdu = CreatePacket(src, dest,
		       ((short) (IPMHLEN + sizeof(A_PDU_TYPE))),
		       pktid, 0, sourcep);

	  num_frags_generated += 1; /* generated one fragment at the source */
           /* send fragment */
	  send_pdu_to_datalink(network_layer_entity, pdu,i);
     }
     else{
	  /* assume no fragmentation at the source */
	  panic("source cannot fragment");

     }


     UpdatePktId(network_layer_entity); 

     /* free the input pdu */
     pdu_free(input_pdu);
     return 0;
}


/* ***************************************************** */
static PDU_TYPE *
CreatePacket(int src, int dest, short pktlen, short pktid,
	     short fragoff, char *sourcep)
{
char *p;
PDU_TYPE *pdu;

     pdu = (PDU_TYPE *)pdu_alloc();
     pdu->type = TYPE_IS_N_PDU;

     pdu->u.n_pdu.ip_len = pktlen;
     pdu->u.n_pdu.ip_id = pktid;
     pdu->u.n_pdu.ip_fragoff = fragoff;
     pdu->u.n_pdu.ip_chksum = 0; /* NO ERROR */
     pdu->u.n_pdu.ip_src = src; 
     pdu->u.n_pdu.ip_dst = dest; 

     p = &(pdu->u.n_pdu.ip_data[0]);
/*     p = (u_char *)malloc((pktlen - IPMHLEN) * sizeof(u_char));  */
     bcopy( sourcep, p, ((pktlen - IPMHLEN) * sizeof(u_char)));
     
     return pdu;
}


/********************************************************************/

/* Out of Order Buffering Data Structure Routines */
/********************************************************************/
/* Creates an ordered list */
static queuet *
CreateOrderedList()
{
     queuet *q;
     q = q_create(); 
     return q;
}
/********************************************************************/


/* Adds pdu to the ordered list based on the 13-bit ordering key 'fragoff'  */
static int
AddToOrderedList(OrderedList_t *q, PDU_TYPE *pdu, int fragoff)
{
OrderedListElement_t *qelt,*qelt2;
PDU_TYPE *p; /* pdu does not have a linking pointer unlike tcp pkt
	      => can't cast pdu directly to q_elt */
short f_off;
short my_off;

if(q->q_head){ /* Non-zero elements in Q=> insert at head or middle */

     /* 13 bit frag offset is the ordering key */
     f_off =
	  ((((PDU_TYPE *)(q->q_head->qe_data))->u.n_pdu.ip_fragoff)
	   & IP_FRAGOFF);
     
     if (fragoff < f_off){  /* smaller than 1st elt => insert before head of list */
	  q_addh(q, (caddr_t)pdu); /* Add to head of queue */
     }
     else {    /* insert in the middle of the list somewhere */
	  for(qelt = q->q_head ; qelt->qe_next; qelt = qelt->qe_next){
	       p = (PDU_TYPE *)qelt->qe_data;
	       f_off = p->u.n_pdu.ip_fragoff & IP_FRAGOFF;
	       my_off =
       ((PDU_TYPE *)(qelt->qe_next->qe_data))->u.n_pdu.ip_fragoff & IP_FRAGOFF;

	       
	       if((f_off <= fragoff) && (fragoff < my_off)) {
		    if (f_off == fragoff){
			 printf("oops, something wrong... ");
			 fflush(stdout);
		    }
		    /* f_off <= fragoff < my_off => insert here */
		    break;
	       } /* if */
	  } /* for */

	  qelt2 = (q_elt *)sim_calloc(1, sizeof(q_elt));
	  qelt2->qe_data = (caddr_t)pdu;
	  qe_adda(q, qelt, qelt2);
     } /* end: insert in middle */
} /* end: insert at head or middle */

else{ /* insert at tail ... */
     q_addt(q,(caddr_t)pdu);
}

return 1;
} /* end routine */

/********************************************************************/
static PDU_TYPE *
TraverseOrderedList(OrderedList_t *q, int *iwantthis)
{
     q_elt *qelt;
     int i, pos = *iwantthis;
     
     if (pos < 0){
	  panic(" illegal position for traversal");
     }

     if (pos == 0){ /* *iwantthis == 0 => start traversal, return first element */
	  qelt = q->q_head;
	  *iwantthis = pos+1;
	  return ((PDU_TYPE *) (qelt->qe_data));
     }
     else{ /* start traversal from the middle */
	  for (i=0 , qelt = q->q_head; i < pos; i++, qelt = qelt->qe_next){
	       if (! qelt->qe_next){
		    if (i+1 == pos) /* you want the end of the list ? */
			 return ((PDU_TYPE *) NULL);
		    else
			 panic(" no available element for this traversal");
	       }
	  }
	  *iwantthis = pos+1;
	   return ((PDU_TYPE *) (qelt->qe_data));
     }
     
}
/********************************************************************/


static PDU_TYPE *
DequeueFromOrderedList(OrderedList_t *q)
{
     return (PDU_TYPE *)q_deq(q);
}

/********************************************************************/
static
FreeOrderedList(OrderedList_t *q)
{
     q_obliterate(q);
}

/********************************************************************/
/* reassembly timer routine: given to students */
/* ***************************************************** */
/* ipftimer  - update time-to-live fields and delete expired fragments */


void ipftimer(NETWORK_LAYER_ENTITY_TYPE *network_layer_entity)
{
int i;
PDU_TYPE *pdu;

for (i=0; i<IP_FQSIZE; i++) {
     struct ipfq *iq = &(network_layer_entity->ipfqt[i]);

     if (iq->ipf_state == IPFF_FREE)
	  continue;
     iq->ipf_ttl -= FTTL_DECR; 

     if (iq->ipf_ttl <= 0) {  /* Timer expired */
	  if (iq->ipf_state == IPFF_BOGUS) {
	       /* resources already gone */
	       iq->ipf_state = IPFF_FREE;
	       iq->nfrags = 0;/* debugging */
	       continue;
	  }
	  if (pdu = (PDU_TYPE *)DequeueFromOrderedList(iq->ipf_q)) {
	       /* reassembly has failed: timeout */
	       pdu_free(pdu);
	  }
	  while (pdu = (PDU_TYPE *)DequeueFromOrderedList(iq->ipf_q)) {
	       pdu_free(pdu);
	  }
	  
	  FreeOrderedList(iq->ipf_q);

	  network_layer_entity->num_packets_timedout++;
	  network_layer_entity->num_partially_reassembled--;
	  dprintf(1, "%8.2f: Node %s: Timer expired for Pkt Id: %d. Dropping... Num Packets Timed out = %d\n", TICKS_TO_USECS(ev_now()), network_layer_entity->cn_name, iq->ipf_id, network_layer_entity->num_packets_timedout);
	  fflush(stdout);
	  
	  iq->ipf_state = IPFF_FREE;
	  iq->nfrags = 0;/* debugging */


     }
}
return;
}
/* ***************************************************** */
