/* ---- DO NOT MODIFY THESE HEADER FILES ----- */
#include "cisePort.h"
#include "sim.h"
#include "component.h"
#include "comptypes.h"
#include "q.h"
#include "list.h"
#include "eventdefs.h"
#include "main.h"
#include "route_activity.h"
#include "sim_tk.h"
#include "network_layer.h"

/****************************************************/
/* ----- DO NOT REMOVE OR MODIFY THIS FUNCTION ---- */
/* NetworkLayerReceive():
   - Can be a Source, Router or Destination
   - Source rcvs from app, sends to DLC
   - Router rcvs from DLC, sends to DLC
   - Destination rcvs from DLC sends to app
   - pdu dropped otherwise
   */

static 
NetworkLayerReceive(NETWORK_LAYER_ENTITY_TYPE *network_layer_entity,
		  GENERIC_LAYER_ENTITY_TYPE *generic_layer_entity,
		  PDU_TYPE *pdu)
{
 if (NetworkFromApplication(generic_layer_entity)) {

      if(SourceEndSystem(network_layer_entity)) {
	   NetworkToDatalink(network_layer_entity, pdu);
      }
      else
	  DropPDU(pdu);
      
 } else if (NetworkFromDatalink(generic_layer_entity)) {

      if(DestinationEndSystem(network_layer_entity)){
	   NetworkToApplication(network_layer_entity, pdu);
      }
      else if (RouterNode(network_layer_entity)){
	   NetworkToDatalink(network_layer_entity, pdu);
      }
      else DropPDU(pdu);
      
 }
   return 0;
}


/********************************************************************/
/* ----- DO NOT REMOVE OR MODIFY THIS FUNCTION ---- */
/*  Fragmentation here 
   - packet can be an A_PDU or an N_PDU, get an N_pdu to work with.
   - Choice of output datalink
   - Fragmentation if required
   - call send_pdu_to_datalink
 */
static
NetworkToDatalink(NETWORK_LAYER_ENTITY_TYPE *network_layer_entity,
		 PDU_TYPE *input_pdu)
{
/*  PDU_TYPE *pdu_to_datalink  = (PDU_TYPE *)pdu_alloc(); */
  int i;
      
  if (input_pdu->type != TYPE_IS_A_PDU && input_pdu->type != TYPE_IS_N_PDU)
       panic("Unknown PDU type: Expect A_PDU or N_PDU\n");

/* 1. Routing part: Choose Datalink (given) */
  i = choose_output_datalink(network_layer_entity);


/* 2. Fragmentation part */
  if (SourceEndSystem(network_layer_entity)){ /* source part (given) */
       FormPacketsAndSendToDLC(network_layer_entity,input_pdu,i);
  }
  else{   /* Router breaks a packet into fragments if necessary */
       FragmentPacketAndSendToDLC(network_layer_entity,input_pdu,i);
  }


  return 0;
}

/* ***************************************************** */
/* ----- DO NOT REMOVE OR MODIFY THIS FUNCTION ---- */
/*  Reassembly here
   - call ipreass
       - returns reassembled pdu (N_PDU) if reassembly complete
       -         NULL otherwise
   - if reassembly complete, form a A_PDU and send to app
 */
static
NetworkToApplication(NETWORK_LAYER_ENTITY_TYPE *network_layer_entity,
		    PDU_TYPE *pdu_from_datalink)
{
  PDU_TYPE *reassembled_pdu;
  PDU_TYPE *pdu_to_app;
  
  if (pdu_from_datalink->type != TYPE_IS_N_PDU) panic("Empty N_pdu\n");

   /* ipreass():  Returns reassembled n_pdu, if complete, NULL otherwise */
  reassembled_pdu = ipreass(network_layer_entity, pdu_from_datalink);

   if (reassembled_pdu) {  /* complete pdu => extract a_pdu */

	pdu_to_app = (PDU_TYPE *)pdu_alloc();
	pdu_to_app->type = TYPE_IS_A_PDU;
	bcopy((char *)&(reassembled_pdu->u.n_pdu.ip_data[0]),
	      (char *)&(pdu_to_app->u.a_pdu),
	      (reassembled_pdu->u.n_pdu.ip_len - IPMHLEN));

	/* -- Send app pdu to app -- */
       send_pdu_to_application(network_layer_entity, pdu_to_app);

       /* -- free reassembled packet --- */
       pdu_free(reassembled_pdu);
	
   }

  return 0;
}


/* ***************************************************** */
/* Fragmentation Subroutines:                            */
/* 1. FragmentPacketAndSendToDLC()    [ You code this ]	 */
/* 2. FormPacketsAndSendToDLC()       [ Given ]		 */
/* 3. CreatePacket()		      [ Given ] 	 */
/* 4. choose_output_datalink()	      [ Given ]          */

/* Useful Macros: 					 */
/* 1. GetMTU()			      [ Given ]		 */
/* ***************************************************** */
/* DO YOUR CODING STARTING HERE */
static
FragmentPacketAndSendToDLC(NETWORK_LAYER_ENTITY_TYPE *network_layer_entity,
			PDU_TYPE *input_pdu,
			int i)
{
     PDU_TYPE *pdu;       /* pdu (fragments) sent to DLC */
     u_char *app_data_ptr;/* pointer within app packet for next fragment data */

     /* -- accounting variables -- */
     int  maxdlen,       /* Max data length of fragment (multiple of 8 bytes)*/
	  tosend,	 /* How much *data*  remains to be sent */
	  dataoffset,	 /* offset within app packet for next fragment */
	  offset_in_datagram; /* fragment offset to be written in new
				 fragment header */

     int N=0; /* count of number of fragments */

     /* -- shorthand variables -- */
     int src; /* Source IP address */
     int dest; /* Destination IP address */
     short pktid; /* Unique Packet Id */
     short mtu; /* subnet MTU */
     short pktlen; /* pklen = header + payload */
     short old_fragoff; /* old fragment offset: from pkt  */

     
     /* sanity check */
     if(input_pdu->type != TYPE_IS_N_PDU)
	  panic("Unknown PDU type: Expect N_PDU \n");

     /* MTU enough: No fragmentation required */
     /* Routing trivial => no processing on header required */
     /* Now send fragment, don't free it  */

     /* Else, if MTU insufficient: Fragmentation required. */

       /* maxdlen: excludes header, multiple of 8 */
       /* dataoffset: offset within app packet for next fragment. 16 bits */
       /* frag offset in datagram: consider old fragment offset too. 16 bits. */
       /* tosend: how much remaining *data* from input_pdu */

       /* Send first N-1 fragments out of N */
	       /* Get a pointer to the data in the app pdu */
	       /* create new packet. packet length is header + data which may
		  not be equal to mtu size. Use same packet id. MF bit must be
		  set. Fix frag offset bits (13 bits) in new packet header */
	       /* call CreatePacket() */

	       /* send fragment to dlc */

	       /* update accounting variables */

       /* Last fragment: if you are fragmenting a fragment with MF bit set,
	  then you need to have MF (more fragments) bit
	  set on the last fragment as well. Else MF is not set. */

       /* send fragment */

       /* final accounting/printing */
       /* debugging hint */
/*
       dprintf(1, "%8.2f: Fragmented Pkt # %d at node %s. %d fragments created. Link %d\n",
		 TICKS_TO_USECS(ev_now()), pktid,
		 network_layer_entity->cn_name, N,i);
	  fflush(stdout);
	  */
     
       /* free the input pdu */
  
     
     return 0;
}
/* ***************************************************** */
/* Reassembly Subroutines:                               */

/* 1. ipreass()			       [ You code this ] */
/* 2. ipfadd()			       [ You code this ] */
/* 3. ipfjoin()			       [ You code this ] */
/* 4. ipfcons()			       [ You code this ] */
/* 5. ipftimer()		       [ Given ] 	 */

/* Out-of-Order Buffering Routines:			 */

/* 1. CreateOrderedList()	       [ Given ] 	 */
/* 2. AddToOrderedList()	       [ Given ] 	 */
/* 3. TraverseOrderedList()	       [ Given ] 	 */
/* 4. DequeueFromOrderedList()	       [ Given ] 	 */
/* 5. FreeOrderedList()	 	       [ Given ] 	 */
/* ***************************************************** */

/* ipreass  - reassembles IP dgram if necessary,
            -  returns N_PDU, if complete;
                      NULL otherwise             */
static PDU_TYPE *
ipreass(NETWORK_LAYER_ENTITY_TYPE *network_layer_entity, PDU_TYPE *pdu)
{
int firstfree = -1;            /* index pointing to free spot in frag. list */
N_PDU_TYPE *pip = &(pdu->u.n_pdu); /* IP pkt pointer */
int i;             
struct ipfq *piq; /* i th frag queue table entry */

/* Is reassembly required ? 
	- fragoff == 0 => first fragment
	- MF not set => no more fragments
	- together => reassembly not required
	- simply return pdu in such a case.
	*/

/* Graphics support (for no reassembly required case). */
if(M.doX)
     call_viewer(network_layer_entity->cn_name, pip->ip_id,
		 (pip->ip_fragoff & IP_FRAGOFF) << 3, pip->ip_len - IPMHLEN);

/* debugging hint */
/*     dprintf(1, "%8.2f: Node %s received dgram. No reassebly required\n",
	    TICKS_TO_USECS(ev_now()), network_layer_entity->cn_name);
     fflush(stdout);
     */


/* Else, reassembly required.
   Look through fragment list (maxsize = IP_FQSIZE)
   for a match/vacant spot */

              /* get i th frag queue table entry */

     /* if it is a free spot, note it down (used later if necessary) */

     /* pkt id or source mismatch => not this one ... */

     /* found a match => add fragment to a partial fragment list: ipfadd() */

     /* debugging hint */
/*
dprintf(1,"%8.2f: Node %s received %d fragments of Pkt id: %d. Last frag payload size = %d bytes\n",
       TICKS_TO_USECS(ev_now()), network_layer_entity->cn_name,
       piq->nfrags, pip->ip_id, (pip->ip_len - IPMHLEN));
fflush(stdout);
*/

     /* check if reassembly complete. If so, join the fragments and
	return reassembled pdu, else return NULL : ipfjoin() */


/* If you are here => you haven't found a match in the fragment lists */
/* Use the free spot to create a new fragment list */

  /* no room in the fragment table => drop pkt */

/* else create an ordered list */
/* use source address and unique pkt id to identify a pkt and its fragments */

/* reset time-to-live (TTL) for this table entry */

/* indicate that this table position taken, and is valid */

/* add the pdu to the list: use ipfadd() */


/* debugging hint */
/*
dprintf(1,"%8.2f: Node %s received %d fragments of Pkt id: %d. Last frag payload size = %d bytes\n",
       TICKS_TO_USECS(ev_now()), network_layer_entity->cn_name,
       network_layer_entity->ipfqt[firstfree].nfrags, pip->ip_id, (pip->ip_len -IPMHLEN));
fflush(stdout);
*/


}

/* ***************************************************** */
/* ipfadd  - add a fragment to an IP fragment queue */

static int
ipfadd(NETWORK_LAYER_ENTITY_TYPE *network_layer_entity,
       struct ipfq *iq, PDU_TYPE *pdu)
{
int fragoff; 		/* 13 bit frag offset in pdu: index to ordered list */
N_PDU_TYPE *pip;	/* IP pkt pointer used as before for convenience */

/* sanity check: valid fragment list ? Else drop pdu */

/* add fragment to ordered list with the 13-bit frag offset as the index */

/* restart time-to-live timer, increment fragment count */

/* Graphics support. DO NOT REMOVE. */
if(M.doX)
     call_viewer(network_layer_entity->cn_name, pip->ip_id,
		 (pip->ip_fragoff & IP_FRAGOFF) << 3, pip->ip_len - IPMHLEN);

return 1;
}

/* ***************************************************** */
/* ipfjoin  - join fragments if all collected */


static PDU_TYPE *
ipfjoin(struct ipfq *iq)
{
PDU_TYPE *tmppdu;   	 /* temporary pdu pointer */
N_PDU_TYPE *pip; 	 /* IP pkt pointer */
int expected_offset,   	 /* expected cumulative fragment offset  */
    packoff;     	 /* fragment offset in the next fragment (16 bit) */
int position=0; 	 /* used for ordered list traversal */

/* sanity check: is this a bogus fragment list ? */

/* Traverse the ordered fragment list. See if we have the whole datagram */
      /* Expected offset must equal fragment offset at every iteration.
         Else, missing fragment, can't join */

     /* increment expected offset by the payload length of this fragment */

/* Here => end of fragment list. Is this the last fragment ? */
   /* Yes, reconstruct and return n_pdu */

}


/* ***************************************************** */
/* ipfcons  - construct a single n_pdu from an IP fragment queue  */

static PDU_TYPE *
ipfcons(struct ipfq *iq)
{
PDU_TYPE *pdu, *tmppdu; /* pdu: reconstructed pdu. tmppdu: temporary ptr */
N_PDU_TYPE *pip;        /* IP pkt pointer */
int off;                /* offset in partially-reassembled packet */
int dlen;		/* payload data length */

/* Allocate and initialize 'pdu' */

/* dequeue pdu from fragment list */


/* bcopy the header: will fix it in the end */

/* Iteratively copy the *data* from fragment (watch your indicies),
   dequeue next pdu. Remember to free tmppdu after you are done with it. */


/* fix header of the reassembled packet (length and fragoff fields) */

/* Release resources in fragment table: ordered list, state variables */

/* return reassembled 'pdu' */
}
/* ***************************************************** */



